--[[
These are Proxmark command definitions.
This file is automatically generated from pm3_cmd.h - DON'T EDIT MANUALLY.
--]]
local __commands = {
CMD_DEVICE_INFO = 0x0000,
CMD_SETUP_WRITE = 0x0001,
CMD_FINISH_WRITE = 0x0003,
CMD_HARDWARE_RESET = 0x0004,
CMD_START_FLASH = 0x0005,
CMD_CHIP_INFO = 0x0006,
CMD_BL_VERSION = 0x0007,
CMD_NACK = 0x00fe,
CMD_ACK = 0x00ff,
CMD_DEBUG_PRINT_STRING = 0x0100,
CMD_DEBUG_PRINT_INTEGERS = 0x0101,
CMD_DEBUG_PRINT_BYTES = 0x0102,
CMD_LCD_RESET = 0x0103,
CMD_LCD = 0x0104,
CMD_BUFF_CLEAR = 0x0105,
CMD_READ_MEM = 0x0106,
CMD_READ_MEM_DOWNLOAD = 0x010A,
CMD_READ_MEM_DOWNLOADED = 0x010B,
CMD_VERSION = 0x0107,
CMD_STATUS = 0x0108,
CMD_PING = 0x0109,
CMD_DOWNLOAD_EML_BIGBUF = 0x0110,
CMD_DOWNLOADED_EML_BIGBUF = 0x0111,
CMD_CAPABILITIES = 0x0112,
CMD_QUIT_SESSION = 0x0113,
CMD_SET_DBGMODE = 0x0114,
CMD_STANDALONE = 0x0115,
CMD_WTX = 0x0116,
CMD_TIA = 0x0117,
CMD_BREAK_LOOP = 0x0118,
CMD_SET_TEAROFF = 0x0119,
CMD_GET_DBGMODE = 0x0120,
CMD_FLASHMEM_WRITE = 0x0121,
CMD_FLASHMEM_WIPE = 0x0122,
CMD_FLASHMEM_DOWNLOAD = 0x0123,
CMD_FLASHMEM_DOWNLOADED = 0x0124,
CMD_FLASHMEM_INFO = 0x0125,
CMD_FLASHMEM_SET_SPIBAUDRATE = 0x0126,
CMD_FLASHMEM_PAGES64K = 0x0127,
CMD_SPIFFS_MOUNT = 0x0130,
CMD_SPIFFS_UNMOUNT = 0x0131,
CMD_SPIFFS_WRITE = 0x0132,
CMD_SPIFFS_APPEND = 0x1132,
CMD_SPIFFS_READ = 0x0133,
CMD_SPIFFS_REMOVE = 0x0134,
CMD_SPIFFS_RM = CMD_SPIFFS_REMOVE,
CMD_SPIFFS_RENAME = 0x0135,
CMD_SPIFFS_MV = CMD_SPIFFS_RENAME,
CMD_SPIFFS_COPY = 0x0136,
CMD_SPIFFS_CP = CMD_SPIFFS_COPY,
CMD_SPIFFS_STAT = 0x0137,
CMD_SPIFFS_FSTAT = 0x0138,
CMD_SPIFFS_INFO = 0x0139,
CMD_SPIFFS_FORMAT = CMD_FLASHMEM_WIPE,
CMD_SPIFFS_WIPE = 0x013A,
CMD_SET_FPGAMODE = 0x013F,
CMD_SPIFFS_PRINT_TREE = 0x2130,
CMD_SPIFFS_GET_TREE = 0x2131,
CMD_SPIFFS_TEST = 0x2132,
CMD_SPIFFS_PRINT_FSINFO = 0x2133,
CMD_SPIFFS_DOWNLOAD = 0x2134,
CMD_SPIFFS_DOWNLOADED = 0x2135,
CMD_SPIFFS_ELOAD = 0x2136,
CMD_SPIFFS_CHECK = 0x3000,
CMD_SMART_RAW = 0x0140,
CMD_SMART_UPGRADE = 0x0141,
CMD_SMART_UPLOAD = 0x0142,
CMD_SMART_ATR = 0x0143,
CMD_SMART_SETBAUD = 0x0144,
CMD_SMART_SETCLOCK = 0x0145,
CMD_USART_RX = 0x0160,
CMD_USART_TX = 0x0161,
CMD_USART_TXRX = 0x0162,
CMD_USART_CONFIG = 0x0163,
CMD_LF_TI_READ = 0x0202,
CMD_LF_TI_WRITE = 0x0203,
CMD_LF_ACQ_RAW_ADC = 0x0205,
CMD_LF_MOD_THEN_ACQ_RAW_ADC = 0x0206,
CMD_DOWNLOAD_BIGBUF = 0x0207,
CMD_DOWNLOADED_BIGBUF = 0x0208,
CMD_LF_UPLOAD_SIM_SAMPLES = 0x0209,
CMD_LF_SIMULATE = 0x020A,
CMD_LF_HID_WATCH = 0x020B,
CMD_LF_HID_SIMULATE = 0x020C,
CMD_LF_SET_DIVISOR = 0x020D,
CMD_LF_SIMULATE_BIDIR = 0x020E,
CMD_SET_ADC_MUX = 0x020F,
CMD_LF_HID_CLONE = 0x0210,
CMD_LF_EM410X_CLONE = 0x0211,
CMD_LF_T55XX_READBL = 0x0214,
CMD_LF_T55XX_WRITEBL = 0x0215,
CMD_LF_T55XX_RESET_READ = 0x0216,
CMD_LF_PCF7931_READ = 0x0217,
CMD_LF_PCF7931_WRITE = 0x0223,
CMD_LF_EM4X_LOGIN = 0x0229,
CMD_LF_EM4X_READWORD = 0x0218,
CMD_LF_EM4X_WRITEWORD = 0x0219,
CMD_LF_EM4X_PROTECTWORD = 0x021B,
CMD_LF_EM4X_BF = 0x022A,
CMD_LF_IO_WATCH = 0x021A,
CMD_LF_EM410X_WATCH = 0x021C,
CMD_LF_EM4X50_INFO = 0x0240,
CMD_LF_EM4X50_WRITE = 0x0241,
CMD_LF_EM4X50_WRITEPWD = 0x0242,
CMD_LF_EM4X50_READ = 0x0243,
CMD_LF_EM4X50_BRUTE = 0x0245,
CMD_LF_EM4X50_LOGIN = 0x0246,
CMD_LF_EM4X50_SIM = 0x0250,
CMD_LF_EM4X50_READER = 0x0251,
CMD_LF_EM4X50_ESET = 0x0252,
CMD_LF_EM4X50_CHK = 0x0253,
CMD_LF_EM4X70_INFO = 0x0260,
CMD_LF_EM4X70_WRITE = 0x0261,
CMD_LF_EM4X70_UNLOCK = 0x0262,
CMD_LF_EM4X70_AUTH = 0x0263,
CMD_LF_EM4X70_SETPIN = 0x0264,
CMD_LF_EM4X70_SETKEY = 0x0265,
CMD_LF_EM4X70_BRUTE = 0x0266,
CMD_LF_SAMPLING_SET_CONFIG = 0x021D,
CMD_LF_FSK_SIMULATE = 0x021E,
CMD_LF_ASK_SIMULATE = 0x021F,
CMD_LF_PSK_SIMULATE = 0x0220,
CMD_LF_NRZ_SIMULATE = 0x0232,
CMD_LF_AWID_WATCH = 0x0221,
CMD_LF_VIKING_CLONE = 0x0222,
CMD_LF_T55XX_WAKEUP = 0x0224,
CMD_LF_COTAG_READ = 0x0225,
CMD_LF_T55XX_SET_CONFIG = 0x0226,
CMD_LF_SAMPLING_PRINT_CONFIG = 0x0227,
CMD_LF_SAMPLING_GET_CONFIG = 0x0228,
CMD_LF_T55XX_CHK_PWDS = 0x0230,
CMD_LF_T55XX_DANGERRAW = 0x0231,
CMD_LF_ZX_READ = 0x0270,
CMD_LF_ZX_WRITE = 0x0271,
CMD_HF_ISO15693_ACQ_RAW_ADC = 0x0300,
CMD_HF_ACQ_RAW_ADC = 0x0301,
CMD_HF_SRI_READ = 0x0303,
CMD_HF_ISO14443B_COMMAND = 0x0305,
CMD_HF_ISO15693_READER = 0x0310,
CMD_HF_ISO15693_SIMULATE = 0x0311,
CMD_HF_ISO15693_SNIFF = 0x0312,
CMD_HF_ISO15693_COMMAND = 0x0313,
CMD_HF_ISO15693_FINDAFI = 0x0315,
CMD_HF_ISO15693_SLIX_ENABLE_PRIVACY = 0x0867,
CMD_HF_ISO15693_SLIX_DISABLE_PRIVACY = 0x0317,
CMD_HF_ISO15693_SLIX_DISABLE_EAS = 0x0318,
CMD_HF_ISO15693_SLIX_ENABLE_EAS = 0x0862,
CMD_HF_ISO15693_SLIX_PASS_PROTECT_AFI = 0x0863,
CMD_HF_ISO15693_SLIX_PASS_PROTECT_EAS = 0x0864,
CMD_HF_ISO15693_SLIX_WRITE_PWD = 0x0865,
CMD_HF_ISO15693_SLIX_PROTECT_PAGE = 0x0868,
CMD_HF_ISO15693_WRITE_AFI = 0x0866,
CMD_HF_TEXKOM_SIMULATE = 0x0320,
CMD_HF_ISO15693_EML_CLEAR = 0x0330,
CMD_HF_ISO15693_EML_SETMEM = 0x0331,
CMD_HF_ISO15693_EML_GETMEM = 0x0332,
CMD_HF_ISO15693_CSETUID = 0x0316,
CMD_HF_ISO15693_CSETUID_V2 = 0x0333,
CMD_LF_SNIFF_RAW_ADC = 0x0360,
CMD_LF_HITAG_SNIFF = 0x0370,
CMD_LF_HITAG_SIMULATE = 0x0371,
CMD_LF_HITAG_READER = 0x0372,
CMD_LF_HITAG2_WRITE = 0x0377,
CMD_LF_HITAG2_CRACK = 0x0378,
CMD_LF_HITAG2_CRACK_2 = 0x0379,
CMD_LF_HITAGS_TEST_TRACES = 0x0367,
CMD_LF_HITAGS_SIMULATE = 0x0368,
CMD_LF_HITAGS_READ = 0x0373,
CMD_LF_HITAGS_WRITE = 0x0375,
CMD_LF_HITAGS_UID = 0x037A,
CMD_LF_HITAGU_READ = 0x037B,
CMD_LF_HITAGU_WRITE = 0x037C,
CMD_LF_HITAGU_SIMULATE = 0x037D,
CMD_LF_HITAGU_UID = 0x037E,
CMD_LF_HITAG_ELOAD = 0x0376,
CMD_HF_ISO14443A_ANTIFUZZ = 0x0380,
CMD_HF_ISO14443B_SIMULATE = 0x0381,
CMD_HF_ISO14443B_SNIFF = 0x0382,
CMD_HF_ISO14443A_SNIFF = 0x0383,
CMD_HF_ISO14443A_SIMULATE = 0x0384,
CMD_HF_ISO14443A_SIM_AID = 0x1420,
CMD_HF_ISO14443A_READER = 0x0385,
CMD_HF_ISO14443A_EMV_SIMULATE = 0x0386,
CMD_HF_LEGIC_SIMULATE = 0x0387,
CMD_HF_LEGIC_READER = 0x0388,
CMD_HF_LEGIC_WRITER = 0x0389,
CMD_HF_EPA_COLLECT_NONCE = 0x038A,
CMD_HF_EPA_REPLAY = 0x038B,
CMD_HF_EPA_PACE_SIMULATE = 0x038C,
CMD_HF_LEGIC_INFO = 0x03BC,
CMD_HF_LEGIC_ESET = 0x03BD,
CMD_HF_ICLASS_READCHECK = 0x038F,
CMD_HF_ICLASS_DUMP = 0x0391,
CMD_HF_ICLASS_SNIFF = 0x0392,
CMD_HF_ICLASS_SIMULATE = 0x0393,
CMD_HF_ICLASS_READER = 0x0394,
CMD_HF_ICLASS_READBL = 0x0396,
CMD_HF_ICLASS_WRITEBL = 0x0397,
CMD_HF_ICLASS_EML_MEMSET = 0x0398,
CMD_HF_ICLASS_CHKKEYS = 0x039A,
CMD_HF_ICLASS_RESTORE = 0x039B,
CMD_HF_ICLASS_CREDIT_EPURSE = 0x039C,
CMD_HF_ICLASS_RECOVER = 0x039D,
CMD_HF_FELICA_SIMULATE = 0x03A0,
CMD_HF_FELICA_SNIFF = 0x03A1,
CMD_HF_FELICA_COMMAND = 0x03A2,
CMD_HF_FELICALITE_DUMP = 0x03AA,
CMD_HF_FELICALITE_SIMULATE = 0x03AB,
CMD_HF_ISO14443A_PRINT_CONFIG = 0x03B0,
CMD_HF_ISO14443A_GET_CONFIG = 0x03B1,
CMD_HF_ISO14443A_SET_CONFIG = 0x03B2,
CMD_HF_ISO14443A_SET_THRESHOLDS = 0x03B8,
CMD_MEASURE_ANTENNA_TUNING = 0x0400,
CMD_MEASURE_ANTENNA_TUNING_HF = 0x0401,
CMD_MEASURE_ANTENNA_TUNING_LF = 0x0402,
CMD_LISTEN_READER_FIELD = 0x0420,
CMD_HF_DROPFIELD = 0x0430,
CMD_FPGA_MAJOR_MODE_OFF = 0x0500,
CMD_HF_MIFARE_EML_MEMCLR = 0x0601,
CMD_HF_MIFARE_EML_MEMSET = 0x0602,
CMD_HF_MIFARE_EML_MEMGET = 0x0603,
CMD_HF_MIFARE_EML_LOAD = 0x0604,
CMD_HF_MIFARE_CSETBL = 0x0605,
CMD_HF_MIFARE_CGETBL = 0x0606,
CMD_HF_MIFARE_CIDENT = 0x0607,
CMD_HF_MIFARE_SIMULATE = 0x0610,
CMD_HF_MIFARE_READER = 0x0611,
CMD_HF_MIFARE_NESTED = 0x0612,
CMD_HF_MIFARE_ACQ_ENCRYPTED_NONCES = 0x0613,
CMD_HF_MIFARE_ACQ_NONCES = 0x0614,
CMD_HF_MIFARE_STATIC_NESTED = 0x0615,
CMD_HF_MIFARE_STATIC_ENC = 0x0616,
CMD_HF_MIFARE_ACQ_STATIC_ENCRYPTED_NONCES = 0x0617,
CMD_HF_MIFARE_READBL = 0x0620,
CMD_HF_MIFARE_READBL_EX = 0x0628,
CMD_HF_MIFAREU_READBL = 0x0720,
CMD_HF_MIFARE_READSC = 0x0621,
CMD_HF_MIFAREU_READCARD = 0x0721,
CMD_HF_MIFARE_WRITEBL = 0x0622,
CMD_HF_MIFARE_WRITEBL_EX = 0x0629,
CMD_HF_MIFARE_VALUE = 0x0627,
CMD_HF_MIFAREU_WRITEBL = 0x0722,
CMD_HF_MIFAREU_WRITEBL_COMPAT = 0x0723,
CMD_HF_MIFARE_CHKKEYS = 0x0623,
CMD_HF_MIFARE_SETMOD = 0x0624,
CMD_HF_MIFARE_CHKKEYS_FAST = 0x0625,
CMD_HF_MIFARE_CHKKEYS_FILE = 0x0626,
CMD_HF_MIFARE_SNIFF = 0x0630,
CMD_HF_MIFARE_MFKEY = 0x0631,
CMD_HF_MIFARE_PERSONALIZE_UID = 0x0632,
CMD_HF_MIFAREUC_AUTH = 0x0724,
CMD_HF_MIFAREULAES_AUTH = 0x0725,
CMD_HF_MIFAREUC_SETPWD = 0x0727,
CMD_HF_DESFIRE_READBL = 0x0728,
CMD_HF_DESFIRE_WRITEBL = 0x0729,
CMD_HF_DESFIRE_AUTH1 = 0x072a,
CMD_HF_DESFIRE_AUTH2 = 0x072b,
CMD_HF_DESFIRE_READER = 0x072c,
CMD_HF_DESFIRE_INFO = 0x072d,
CMD_HF_DESFIRE_COMMAND = 0x072e,
CMD_HF_MIFARE_NACK_DETECT = 0x0730,
CMD_HF_MIFARE_STATIC_NONCE = 0x0731,
CMD_HF_MIFARE_STATIC_ENCRYPTED_NONCE = 0x0732,
CMD_HF_MFU_OTP_TEAROFF = 0x0740,
CMD_HF_MFU_COUNTER_TEAROFF = 0x0741,
CMD_HF_SNIFF = 0x0800,
CMD_HF_PLOT = 0x0801,
CMD_FPGAMEM_DOWNLOAD = 0x0802,
CMD_FPGAMEM_DOWNLOADED = 0x0803,
CMD_HF_THINFILM_READ = 0x0810,
CMD_HF_THINFILM_SIMULATE = 0x0811,
CMD_HF_CRYPTORF_SIM = 0x0820,
CMD_HF_MIFARE_GEN3UID = 0x0850,
CMD_HF_MIFARE_GEN3BLK = 0x0851,
CMD_HF_MIFARE_GEN3FREEZ = 0x0852,
CMD_HF_MIFARE_G4_RDBL = 0x0860,
CMD_HF_MIFARE_G4_WRBL = 0x0861,
CMD_HF_MIFARE_G4_GDM_RDBL = 0x0870,
CMD_HF_MIFARE_G4_GDM_WRBL = 0x0871,
CMD_HF_SAM_PICOPASS = 0x0900,
CMD_HF_SAM_SEOS = 0x0901,
CMD_HF_SAM_MFC = 0x0902,
CMD_UNKNOWN = 0xFFFF,
}
return __commands
